unit MAINMENU;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, DBGrids,
  ExtCtrls, Buttons, Menus, SQLite3Conn, SQLDB, DB, IniFiles, IDataExchange,
  FormDatenEinlesen, ShellApi, Windows, Types, Grids;

type
  { TMainMenuForm }
  TMainMenuForm = class(TForm, IDataExch)
    BtnExit: TButton;
    BtnDatenEinlesen: TButton;
    Button2: TButton;
    Button3: TButton;
    Button4: TButton;
    DataSourceListMainMenu: TDataSource;
    DBGrid1: TDBGrid;
    EditDatabasePath: TEdit;
    Label1: TLabel;
    LabelMessAnz: TLabel;
    LabelMessdaten: TLabel;
    MemoDS: TMemo;
    SQLQueryListMainMenu: TSQLQuery;
    StaticText1: TStaticText;
    StaticText3: TStaticText;
    procedure BtnDatenEinlesenClick(Sender: TObject);
    procedure BtnExitClick(Sender: TObject);
    procedure DBGrid1CellClick(Column: TColumn);
    procedure DBGrid1DrawColumnCell(Sender: TObject; const Rect: TRect;
      DataCol: Integer; Column: TColumn; State: TGridDrawState);
    procedure MemoDSChange(Sender: TObject);
    procedure DBGrid1KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
  private
    SQLite3Connection: TSQLite3Connection;
    SQLTransaction: TSQLTransaction;
    DatabaseFile: string;
    DatabaseClosed: Boolean;
    procedure InitializeDatabase;
    procedure LoadDatabasePath;
    procedure SaveDatabasePath;
    procedure CreateTables;
    procedure UpdateDatabasePathDisplay;
    procedure CheckSQLite3DLL;
    procedure VerifyDatabaseStructure;
    procedure CheckForMeasurementData;
    procedure PromptForDatabase;
    procedure CloseDatabase;
    procedure LoadDataSet;
    procedure ConfigureDBGridColumns;
    procedure DBGridDblClick(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure UpdateAndSelect(ADataSetID: Integer);
  end;

var
  MainMenuForm: TMainMenuForm;

implementation

{$R *.lfm}

constructor TMainMenuForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SQLite3Connection := TSQLite3Connection.Create(Self);
  SQLTransaction := TSQLTransaction.Create(Self);
  SQLite3Connection.Transaction := SQLTransaction;
  DatabaseClosed := False;
  SQLQueryListMainMenu.DataBase := SQLite3Connection;
  SQLQueryListMainMenu.Transaction := SQLTransaction;
  DataSourceListMainMenu.DataSet := SQLQueryListMainMenu;
  DBGrid1.DataSource := DataSourceListMainMenu;
  DBGrid1.OnDblClick := @DBGridDblClick;
  DBGrid1.OnKeyDown := @DBGrid1KeyDown;
  KeyPreview := True;
  CheckSQLite3DLL;
  InitializeDatabase;
  LoadDataSet;
end;

destructor TMainMenuForm.Destroy;
begin
  CloseDatabase;
  inherited Destroy;
end;

procedure TMainMenuForm.LoadDataSet;
begin
  SQLQueryListMainMenu.Close;
  SQLQueryListMainMenu.SQL.Text := 'SELECT DataSet.ID, DataSet.CreationDate, DataSet.Name, DataSet.Comment, UnsortedValues.Value FROM DataSet LEFT JOIN UnsortedValues ON DataSet.ID = UnsortedValues.DataSetID ORDER BY DataSet.CreationDate DESC';
  SQLQueryListMainMenu.Open;

  if SQLQueryListMainMenu.IsEmpty then
  begin
    DBGrid1.Columns.Clear;
    with DBGrid1.Columns.Add do
    begin
      Title.Caption := 'KEINE DATENSÄTZE VORHANDEN! Bitte Neue unsortierte Daten einlesen.';
      Width := DBGrid1.ClientWidth;
    end;
  end
  else
  begin
    ConfigureDBGridColumns;
    DBGrid1.Refresh;
    DBGrid1.SelectedIndex := 0;
    CheckForMeasurementData;
  end;
end;

procedure TMainMenuForm.CloseDatabase;
begin
  if not DatabaseClosed then
  begin
    try
      if SQLTransaction.Active then
        SQLTransaction.Rollback;
      if SQLite3Connection.Connected then
        SQLite3Connection.Close;
    except
      on E: Exception do
        ShowMessage('Fehler beim Schließen der Datenbankverbindung: ' + E.Message);
    end;
    DatabaseClosed := True;
  end;
end;

procedure TMainMenuForm.CheckSQLite3DLL;
const
  DLLFileName = 'sqlite3.dll';
begin
  if not FileExists(DLLFileName) then
  begin
    ShowMessage('Fehlende Datei: ' + DLLFileName + '. Bitte sicherstellen, dass die Datei im selben Verzeichnis wie die Anwendung vorhanden ist.');
    Application.Terminate;
  end;
end;

procedure TMainMenuForm.LoadDatabasePath;
var
  IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(ExtractFilePath(Application.ExeName) + 'config.ini');
  try
    DatabaseFile := IniFile.ReadString('Database', 'Path', '');
  finally
    IniFile.Free;
  end;
end;

procedure TMainMenuForm.SaveDatabasePath;
var
  IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(ExtractFilePath(Application.ExeName) + 'config.ini');
  try
    IniFile.WriteString('Database', 'Path', DatabaseFile);
  finally
    IniFile.Free;
  end;
end;

procedure TMainMenuForm.UpdateDatabasePathDisplay;
begin
  EditDatabasePath.Text := DatabaseFile;
end;

procedure TMainMenuForm.PromptForDatabase;
var
  Response: Integer;
begin
  Response := MessageDlg('Möchten Sie eine bereits existierende Datenbank verwenden?', mtConfirmation, [mbYes, mbNo], 0);
  if Response = mrYes then
  begin
    with TOpenDialog.Create(Self) do
    try
      Filter := 'SQLite Database (*.db)|*.db';
      if Execute then
      begin
        DatabaseFile := FileName;
        SaveDatabasePath;
      end
      else
        Application.Terminate;
    finally
      Free;
    end;
    try
      SQLite3Connection.DatabaseName := DatabaseFile;
      SQLite3Connection.Open;
      VerifyDatabaseStructure;
    except
      on E: Exception do
      begin
        ShowMessage('Fehlerhafte Datenbankstruktur: ' + E.Message);
        DatabaseFile := '';
      end;
    end;
  end
  else
  begin
    with TSaveDialog.Create(Self) do
    try
      Filter := 'SQLite Database (*.db)|*.db';
      if Execute then
      begin
        DatabaseFile := FileName;
        SaveDatabasePath;
      end
      else
        Application.Terminate;
    finally
      Free;
    end;
  end;
end;

procedure TMainMenuForm.VerifyDatabaseStructure;
var
  SQLQuery: TSQLQuery;
begin
  SQLQuery := TSQLQuery.Create(nil);
  try
    SQLQuery.DataBase := SQLite3Connection;
    SQLQuery.Transaction := SQLTransaction;
    SQLQuery.SQL.Text := 'SELECT name FROM sqlite_master WHERE type=''table'' AND name IN (''DataSet'', ''UnsortedValues'', ''Measurements'')';
    SQLQuery.Open;
    if SQLQuery.RecordCount < 3 then
      raise Exception.Create('Die Datenbank enthält nicht die erforderlichen Tabellen.');
  finally
    SQLQuery.Free;
  end;
end;

procedure TMainMenuForm.BtnExitClick(Sender: TObject);
var
  Response: Integer;
begin
  Response := MessageDlg('Sind Sie sicher, dass Sie die Anwendung beenden möchten?', mtConfirmation, [mbYes, mbNo], 0);
  if Response = mrYes then
  begin
    CloseDatabase;
    Close;
  end;
end;

procedure TMainMenuForm.DBGrid1CellClick(Column: TColumn);
begin
  CheckForMeasurementData;
end;

procedure TMainMenuForm.DBGrid1KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if (Key in [VK_UP, VK_DOWN, VK_RETURN]) then
  begin
    CheckForMeasurementData;
  end;
end;

procedure TMainMenuForm.BtnDatenEinlesenClick(Sender: TObject);
begin
  try
    if not Assigned(frmDatenEinlesen) then
    begin
      frmDatenEinlesen := TFormDatenEinlesen.Create(Self);
      frmDatenEinlesen.SQLite3Connection := SQLite3Connection;
      frmDatenEinlesen.SQLTransaction := SQLTransaction;
      frmDatenEinlesen.DataExchange := Self;
    end;
    frmDatenEinlesen.ShowModal;
  except
    on E: Exception do
    begin
      ShowMessage('Exception in BtnDatenEinlesenClick: ' + E.Message);
    end;
  end;
  frmDatenEinlesen.Free;
  frmDatenEinlesen := nil;
end;

procedure TMainMenuForm.CheckForMeasurementData;
var
  CommentField: TField;
  DataSetID: Integer;
  MeasurementCount: Integer;
  SQLQuery: TSQLQuery;
begin
  if not Assigned(MemoDS) then Exit;
  if not Assigned(DBGrid1.DataSource.DataSet) then Exit;
  CommentField := DBGrid1.DataSource.DataSet.FieldByName('Comment');
  if Assigned(CommentField) then
  begin
    MemoDS.Text := CommentField.AsString;
  end;

  DataSetID := DBGrid1.DataSource.DataSet.FieldByName('ID').AsInteger;
  SQLQuery := TSQLQuery.Create(nil);
  try
    SQLQuery.DataBase := SQLite3Connection;
    SQLQuery.SQL.Text := 'SELECT COUNT(*) AS Count FROM Measurements WHERE DataSetID = :DataSetID';
    SQLQuery.ParamByName('DataSetID').AsInteger := DataSetID;
    SQLQuery.Open;
    MeasurementCount := SQLQuery.Fields[0].AsInteger;
    if MeasurementCount > 0 then
    begin
      LabelMessdaten.Caption := 'Vorhanden:';
      LabelMessdaten.Font.Color := clGreen;
      LabelMessAnz.Caption := IntToStr(MeasurementCount);
      LabelMessAnz.Font.Color := clGreen;
    end
    else
    begin
      LabelMessdaten.Caption := 'Keine Daten:';
      LabelMessdaten.Font.Color := clRed;
      LabelMessAnz.Caption := '0';
      LabelMessAnz.Font.Color := clRed;
    end;
  finally
    SQLQuery.Free;
  end;
end;

procedure TMainMenuForm.UpdateAndSelect(ADataSetID: Integer);
begin
  SQLQueryListMainMenu.Close;
  SQLQueryListMainMenu.Open;
  DBGrid1.Refresh;
  if DataSourceListMainMenu.DataSet.Locate('ID', ADataSetID, []) then
  begin
    DBGrid1.SelectedRows.CurrentRowSelected := True;
    CheckForMeasurementData;
  end;
end;

procedure TMainMenuForm.InitializeDatabase;
begin
  LoadDatabasePath;
  DataSourceListMainMenu.DataSet := SQLQueryListMainMenu;
  DBGrid1.DataSource := DataSourceListMainMenu;
  DBGrid1.Options := DBGrid1.Options + [dgDisplayMemoText];
  if DatabaseFile = '' then
  begin
    PromptForDatabase;
  end;
  SQLite3Connection.DatabaseName := DatabaseFile;
  try
    SQLite3Connection.Open;
    if not SQLTransaction.Active then
    begin
      SQLTransaction.StartTransaction;
    end;
    CreateTables;
    SQLTransaction.Commit;
  except
    on E: Exception do
    begin
      ShowMessage('Fehler beim Verbinden mit der Datenbank: ' + E.Message);
      if SQLTransaction.Active then
      begin
        SQLTransaction.Rollback;
      end;
    end;
  end;
  UpdateDatabasePathDisplay;
  ConfigureDBGridColumns;
end;

procedure TMainMenuForm.ConfigureDBGridColumns;
const
  TotalGridWidth = 680;
var
  WidthPercentages: array[0..2] of Integer;
  TotalPercent, i: Integer;
begin
  DBGrid1.Columns.Clear;
  WidthPercentages[0] := 30;
  WidthPercentages[1] := 60;
  WidthPercentages[2] := 10;
  TotalPercent := 0;
  for i := 0 to High(WidthPercentages) do
  begin
    TotalPercent := TotalPercent + WidthPercentages[i];
  end;
  with DBGrid1.Columns.Add do
  begin
    FieldName := 'CreationDate';
    Title.Caption := 'Datum';
    Width := (WidthPercentages[0] * TotalGridWidth) div TotalPercent;
  end;
  with DBGrid1.Columns.Add do
  begin
    FieldName := 'Name';
    Title.Caption := 'Name';
    Width := (WidthPercentages[1] * TotalGridWidth) div TotalPercent;
  end;
  with DBGrid1.Columns.Add do
  begin
    FieldName := 'Value';
    Title.Caption := 'Inhalt';
    Width := (WidthPercentages[2] * TotalGridWidth) div TotalPercent;
  end;
end;

procedure TMainMenuForm.DBGrid1DrawColumnCell(Sender: TObject; const Rect: TRect;
  DataCol: Integer; Column: TColumn; State: TGridDrawState);
var
  CellIcon: TIcon;
  IconRect: TRect;
  IconSize: Integer;
begin
  if (Column.FieldName = 'Value') then
  begin
    CellIcon := TIcon.Create;
    try
      CellIcon.LoadFromFile(ExtractFilePath(Application.ExeName) + 'icons\TXT.ico');
      IconSize := Min(Rect.Height, Rect.Width);
      IconRect.Left := Rect.Left + (Rect.Width - IconSize) div 2;
      IconRect.Top := Rect.Top + (Rect.Height - IconSize) div 2;
      IconRect.Right := IconRect.Left + IconSize;
      IconRect.Bottom := IconRect.Top + IconSize;
      DBGrid1.Canvas.FillRect(Rect);
      DBGrid1.Canvas.StretchDraw(IconRect, CellIcon);
    finally
      CellIcon.Free;
    end;
  end
  else
  begin
    DBGrid1.DefaultDrawColumnCell(Rect, DataCol, Column, State);
  end;
end;

procedure TMainMenuForm.MemoDSChange(Sender: TObject);
var
  SelectedDataSetID: Integer;
  SQLUpdate: TSQLQuery;
begin
  if not Assigned(DBGrid1.DataSource.DataSet) or (DBGrid1.DataSource.DataSet.IsEmpty) then Exit;
  SelectedDataSetID := DBGrid1.DataSource.DataSet.FieldByName('ID').AsInteger;
  SQLUpdate := TSQLQuery.Create(nil);
  try
    SQLUpdate.DataBase := SQLite3Connection;
    SQLUpdate.SQL.Text := 'UPDATE DataSet SET Comment = :Comment WHERE ID = :ID';
    SQLUpdate.ParamByName('Comment').AsString := MemoDS.Text;
    SQLUpdate.ParamByName('ID').AsInteger := SelectedDataSetID;
    SQLUpdate.ExecSQL;
    SQLTransaction.Commit;
  finally
    SQLUpdate.Free;
  end;
  SQLQueryListMainMenu.Close;
  SQLQueryListMainMenu.Open;
  if not DBGrid1.DataSource.DataSet.Locate('ID', SelectedDataSetID, []) then
  begin
    ShowMessage('Der zuvor ausgewählte Datensatz kann nicht gefunden werden.');
  end;
end;

procedure TMainMenuForm.DBGridDblClick(Sender: TObject);
var
  BlobStream: TStream;
  TempFileName: String;
  FileStream: TFileStream;
begin
  if DBGrid1.SelectedField.FieldName = 'Value' then
  begin
    BlobStream := SQLQueryListMainMenu.CreateBlobStream(DBGrid1.SelectedField, bmRead);
    try
      TempFileName := GetTempDir + 'temp_blob_file.txt';
      FileStream := TFileStream.Create(TempFileName, fmCreate);
      try
        FileStream.CopyFrom(BlobStream, BlobStream.Size);
      finally
        FileStream.Free;
      end;
      ShellExecute(0, 'open', PChar(TempFileName), nil, nil, SW_SHOW);
    finally
      BlobStream.Free;
    end;
  end;
end;

procedure TMainMenuForm.CreateTables;
begin
  with SQLite3Connection do
  begin
    ExecuteDirect('CREATE TABLE IF NOT EXISTS DataSet (' +
                  'ID INTEGER PRIMARY KEY AUTOINCREMENT, ' +
                  'CreationDate DATETIME, ' +
                  'Name TEXT, ' +
                  'Comment TEXT)');
    ExecuteDirect('CREATE TABLE IF NOT EXISTS UnsortedValues (' +
                  'DataSetID INTEGER, ' +
                  'Value BLOB, ' +
                  'FOREIGN KEY (DataSetID) REFERENCES DataSet(ID))');
    ExecuteDirect('CREATE TABLE IF NOT EXISTS Measurements (' +
                  'ID INTEGER PRIMARY KEY AUTOINCREMENT, ' +
                  'DataSetID INTEGER, ' +
                  'MeasurementDate DATETIME, ' +
                  'AlgorithmName TEXT, ' +
                  'M1 REAL, M2 REAL, M3 REAL, M4 REAL, M5 REAL, ' +
                  'FOREIGN KEY (DataSetID) REFERENCES DataSet(ID))');
  end;
end;

end.

